/*
 * Copyright (c) 2003 Red Hat, Inc. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Akira TAGOH <tagoh@redhat.com>
 */

#include <string.h>
#include "ghost.h"
#include "ialloc.h"
#include "iddict.h"
#include "idict.h"
#include "iname.h"
#include "oper.h"
#include "store.h"

#ifdef HAVE_FONTCONFIG
#include <fontconfig/fontconfig.h>
#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_SFNT_NAMES_H
#include FT_TRUETYPE_IDS_H

/* - zfcbuilddict - */
/* <dict> .fcbuilddict <dict> */
private int
zfcbuilddict (i_ctx_t *i_ctx_p)
{
	os_ptr op = osp;
	FcFontSet *fontset;
	FcPattern *font;
	FcResult result;
	FcChar8 *fc_family, *fc_style, *fc_file;
	int outline;
	int i, j, k, plen, flen, retval;
	uint dictlen, dictmaxlen;
	char *ext;
	byte *psname = NULL, *file;
	ref *value, *tmpary;
	ref rarray, rpsname, rfile;

	/* check the stack whether or not it's a dictionary and is writable */
	check_type (*op, t_dictionary);
	check_dict_write (*op);

	fontset = FcConfigGetFonts (NULL, FcSetSystem);
	if (fontset == NULL) {
		return 0;
	}
	for (i = 0; i < fontset->nfont; i++) {
		font = fontset->fonts[i];
		result = FcPatternGetBool (font, FC_OUTLINE, 0, &outline);
		if (!outline) {
			//dlprintf ("DEBUG: FC_OUTLINE mismatch\n");
			continue;
		}
		result = FcPatternGetString (font, FC_FILE, 0, &fc_file);
		if (result != FcResultMatch || fc_file == NULL) {
			dlprintf ("DEBUG: FC_FILE mismatch\n");
			continue;
		}
		/* FIXME: need to find out the correct way to know the font type. */
		/* pick TrueType font only up here. maybe it's wrong, though */
		ext = rindex (fc_file, '.');
		if (!ext) {
			//dlprintf ("DEBUG: file extension is missing\n");
			continue;
		} else if (!strcasecmp (ext, ".ttf") ||
			   !strcasecmp (ext, ".ttc") ||
			   !strcasecmp (ext, ".otf") ||
			   !strcasecmp (ext, ".font")) {
			//dlprintf ("DEBUG: Fontmap doesn't support OpenType/TrueType font\n");
			continue;
		}

		result = FcPatternGetString (font, FC_FAMILY, 0, &fc_family);
		if (result != FcResultMatch || fc_family == NULL) {
			//dlprintf ("DEBUG: FC_FAMILY mismatch\n");
			continue;
		}
		result = FcPatternGetString (font, FC_STYLE, 0, &fc_style);
		if (result != FcResultMatch || fc_style == NULL) {
			//dlprintf ("DEBUG: FC_STYLE mismatch\n");
			continue;
		}

		/* FIXME: probably generating the psname from the family is a stupid way. */
		plen = strlen (fc_family) + strlen (fc_style);
		psname = ialloc_string (plen + 1, "fcbuilddict_psname");
		if (psname == 0)
			return_error (e_VMerror);
		for (j = 0, k = 0; fc_family[j]; j++) {
			/* ignore the space */
			if (fc_family[j] != ' ')
				psname[k++] = fc_family[j];
		}
		psname[k++] = '-';
		for (j = 0; fc_style[j]; j++) {
			/* ignore the space */
			if (fc_style[j] != ' ')
				psname[k++] = fc_style[j];
		}
		psname[k] = 0;

		if (psname[0] == 0) {
			//dlprintf ("DEBUG: no psname is provided.\n");
			continue;
		}

		/* make a string object for font filename */
		flen = strlen (fc_file);
		file = ialloc_string (flen + 1, "fcbuilddict_fontname");
		if (file == 0) {
			return_error (e_VMerror);
		}
		strncpy (file, fc_file, flen);
		file[flen] = 0;
		make_string (&rpsname, a_all | icurrent_space, plen, psname);
		make_string (&rfile, a_all | icurrent_space, flen, file);

		/* grow the dictionary if needed */
		dictlen = dict_length (op);
		dictmaxlen = dict_maxlength (op);

		if (dictlen == dictmaxlen) {
			//dlprintf1 ("DEBUG: growing the dictionary: max: %d\n", dictmaxlen);
			if (!dict_find (op, &rpsname, &value)) {
				dictmaxlen *= 3;
				dictmaxlen /= 2;
				dictmaxlen++;
#if arch_sizeof_int < arch_sizeof_long
				if (dictmaxlen > max_uint)
					return_error (e_rangecheck);
#else
				if (dictmaxlen < 0)
					return_error (e_rangecheck);
#endif
				retval = idict_resize (op, dictmaxlen);
				if (retval < 0) {
					return retval;
				}
			}
		}

		retval = name_from_string (&rpsname, &rpsname);
		if (retval < 0) {
			dlprintf ("DEBUG: failed to convert string to name.\n");
			return retval;
		}

		//dlprintf2 ("DEBUG: psname: %s, file: %s\n", psname, fc_file);

		/* make an array */
		if (ialloc_ref_array (&rarray, a_all, 1, "fcbuilddict_array") < 0) {
			return_error (e_VMerror);
		}
		ref_assign_old (&rarray, rarray.value.refs, &rfile, "fcbuilddict_put_file");

		retval = idict_put (op, &rpsname, &rarray);
		if (retval < 0) {
			dlprintf ("DEBUG: failed to put the value to the dictionary.\n");
			return retval;
		}
	}

	return 0;
}

/* - zfcbuildciddict - */
/* <dict> .fcbuildciddict <dict> */
private int
zfcbuildciddict (i_ctx_t *i_ctx_p)
{
	os_ptr op = osp;
	FcFontSet *fontset;
	FcPattern *font;
	FcResult result;
	FcChar8 *fc_family, *fc_style, *fc_file;
	FcChar32 fc_index;
	FT_Library ftlib;
	FT_Face face = NULL;
	FT_UInt csfnt, isfnt;
	FT_SfntName sfnt;
	FT_Error fterror;
	int outline;
	int i, j, plen, flen, retval;
	uint dictlen, dictmaxlen;
	char *ext;
	byte *psname = NULL, *file;
	ref *value, *tmpary;
	ref rarray, rpsname, rfile, ridx;

	/* check the stack whether or not it's a dictionary and is writable */
	check_type (*op, t_dictionary);
	check_dict_write (*op);

	if (FT_Init_FreeType (&ftlib)) {
		return 0;
	}
	fontset = FcConfigGetFonts (NULL, FcSetSystem);
	if (fontset == NULL) {
		return 0;
	}
	for (i = 0; i < fontset->nfont; i++) {
		font = fontset->fonts[i];
		result = FcPatternGetBool (font, FC_OUTLINE, 0, &outline);
		if (!outline) {
			//dlprintf ("DEBUG: FC_OUTLINE mismatch\n");
			continue;
		}
		result = FcPatternGetString (font, FC_FILE, 0, &fc_file);
		if (result != FcResultMatch || fc_file == NULL) {
			//dlprintf ("DEBUG: FC_FILE mismatch\n");
			continue;
		}
		/* FIXME: need to find out the correct way to know the font type. */
		/* pick TrueType font only up here */
		ext = rindex (fc_file, '.');
		if (!ext) {
			//dlprintf ("DEBUG: file extention is missing\n");
			continue;
		} else if (strcasecmp (ext, ".ttf") &&
			   strcasecmp (ext, ".ttc") &&
			   strcasecmp (ext, ".otf")) {
			//dlprintf ("DEBUG: neither TrueType font nor OpenType font\n");
			continue;
		}

		result = FcPatternGetString (font, FC_FAMILY, 0, &fc_family);
		if (result != FcResultMatch || fc_family == NULL) {
			//dlprintf ("DEBUG: FC_FAMILY mismatch\n");
			continue;
		}
		result = FcPatternGetString (font, FC_STYLE, 0, &fc_style);
		if (result != FcResultMatch || fc_style == NULL) {
			//dlprintf ("DEBUG: FC_STYLE mismatch\n");
			continue;
		}
		result = FcPatternGetInteger (font, FC_INDEX, 0, &fc_index);
		if (result != FcResultMatch) {
			//dlprintf ("DEBUG: FC_INDEX mismatch\n");
			continue;
		}
		/* if needed, destroy the face */
		if (face) {
			FT_Done_Face (face);
			face = NULL;
		}
		if (FT_New_Face (ftlib, fc_file, fc_index, &face)) {
			dlprintf ("DEBUG: cannot create the face\n");
			continue;
		}
		/* ghostscript doesn't need the bold font and the italic font. */
		if (strncmp (fc_style, "Regular", 7) != 0 &&
		    strncmp (fc_style, "Roman", 5) != 0) {
			//dlprintf1 ("DEBUG: font has unnecessary style. %s\n", fc_style);
			continue;
		}
		/* make a string object for PostScript name */
		csfnt = FT_Get_Sfnt_Name_Count (face);
		psname = NULL;
		for (isfnt = 0; isfnt < csfnt; isfnt++) {
			fterror = FT_Get_Sfnt_Name (face, isfnt, &sfnt);
			if (fterror != FT_Err_Ok) {
				break;
			}
			/* get the PostScript name from English name id */
			if (((sfnt.platform_id == 1 && sfnt.encoding_id == 0) ||
			     (sfnt.platform_id == 3 && sfnt.encoding_id == 1)) &&
			    sfnt.name_id == TT_NAME_ID_PS_NAME) {
				/* validate the ps name */
				for (j = 0; j < sfnt.string_len; j++) {
					if (sfnt.string[j] == 0 || sfnt.string[j] == ' ') {
						/* probably Sfnt has invalid psname */
						/* so going to make it from the family name */
						plen = strlen (fc_family);
						psname = ialloc_string (plen + 1, "fcbuildciddict_psname");
						if (psname == 0)
							return_error (e_VMerror);
						for (j = 0; fc_family[j]; j++) {
							if (fc_family[j] == ' ')
								psname[j] = '-';
							else
								psname[j] = fc_family[j];
						}
						psname[j] = 0;
						break;
					}
				}
				if (!psname) {
					plen = sfnt.string_len;
					psname = ialloc_string (plen + 1, "fcbuildciddict_psname");
					if (psname == 0)
						return_error (e_VMerror);
					memcpy (psname, sfnt.string, sfnt.string_len);
					psname[sfnt.string_len] = 0;
					break;
				} else {
					break;
				}
			}
		}
		/* if fails to get the psname, stop to process */
		if (psname == NULL) {
			dlprintf ("DEBUG: no psname was provided.\n");
			continue;
		}

		/* make a string object for TrueType font filename */
		flen = strlen (fc_file);
		file = ialloc_string (flen + 1, "fcbuildciddict_ttfname");
		if (file == 0) {
			return_error (e_VMerror);
		}
		strncpy (file, fc_file, flen);
		file[flen] = 0;
		make_string (&rpsname, a_all | icurrent_space, plen, psname);
		make_string (&rfile, a_all | icurrent_space, flen, file);
		make_int (&ridx, fc_index + 1);

		/* grow the dictionary if needed */
		dictlen = dict_length (op);
		dictmaxlen = dict_maxlength (op);

		if (dictlen == dictmaxlen) {
			//dlprintf1 ("DEBUG: growing the dictionary: max: %d\n", dictmaxlen);
			if (!dict_find (op, &rpsname, &value)) {
				dictmaxlen *= 3;
				dictmaxlen /= 2;
				dictmaxlen++;
#if arch_sizeof_int < arch_sizeof_long
				if (dictmaxlen > max_uint)
					return_error (e_rangecheck);
#else
				if (dictmaxlen < 0)
					return_error (e_rangecheck);
#endif
				retval = idict_resize (op, dictmaxlen);
				if (retval < 0) {
					return retval;
				}
			}
		}

		retval = name_from_string (&rpsname, &rpsname);
		if (retval < 0) {
			dlprintf ("DEBUG: failed to convert string to name.\n");
			return retval;
		}

		//dlprintf3 ("DEBUG: psname: %s, file: %s, idx: %d\n", psname, fc_file, fc_index);
		/* make an array */
		if (ialloc_ref_array (&rarray, a_all, 2, "fcbuildciddict_array") < 0) {
			return_error (e_VMerror);
		}
		ref_assign_old (&rarray, rarray.value.refs, &rfile, "fcbuildciddict_put_file");
		tmpary = rarray.value.refs + 1;
		ref_assign_old (&rarray, tmpary, &ridx, "fcbuildciddict_put_idx");

		retval = idict_put (op, &rpsname, &rarray);
		if (retval < 0) {
			dlprintf ("DEBUG: failed to put the value to the dictionary.\n");
			return retval;
		}
	}

	if (face)
		FT_Done_Face (face);
	FT_Done_FreeType (ftlib);

	return 0;
}
#else
private int
zfcbuilddict (i_ctx_t *i_ctx_p)
{
	return 0;
}

private int
zfcbuildciddict (i_ctx_t *i_ctx_p)
{
	return 0;
}
#endif /* HAVE_FONTCONFIG */

/* ------ Initialization procedure ------ */
const op_def zfontconfig_op_defs[] = {
	{"1.fcbuilddict", zfcbuilddict},
	{"1.fcbuildciddict", zfcbuildciddict},
	op_def_end (0)
};
